import laspy


def get_nb_points(path):
    """Get number of points in a las file"""
    with laspy.open(path) as f:
        nb_points = f.header.point_count

    return nb_points


def get_2d_bounding_box(path):
    """Get bbox for a las file (x, y only)"""
    with laspy.open(path) as f:
        mins = f.header.mins
        maxs = f.header.maxs

    return mins[:2], maxs[:2]


def get_classification_values(path):
    """Get the set of classification values existing in a las file"""
    with laspy.open(path) as f:
        las = f.read()
        classes = set(las.classification)

    return classes



from pathlib import Path

import geopandas as gpd
from pyproj import CRS
from shapely.geometry import LineString

from las_digital_models.extract_stat_from_raster.vectors.clip_geometry import (
    clip_lines_by_polygons,
)

TEST_PATH = Path(__file__).resolve().parent.parent

TMP_PATH = TEST_PATH / "tmp"

DATA_DIR = TEST_PATH / "data" / "bridge" / "input_operators"
DATA_LINES = DATA_DIR / "lignes_contraintes/test_constraint_lines_3D_OE.geojson"
DATA_LINES_OUTSIDE = DATA_DIR / "lignes_contraintes/test_constraint_lines_3D_OE_outside.geojson"
DATA_BRIDGES = DATA_DIR / "tabliers/test_tabliers_pont_OE.geojson"


def test_create_hydro_vector_mask_default():
    # Parameters
    crs = CRS.from_epsg(2154)
    lines_gdf = gpd.read_file(DATA_LINES)
    bridges_gdf = gpd.read_file(DATA_BRIDGES)

    gdf = clip_lines_by_polygons(lines_gdf, bridges_gdf)

    assert not gdf.empty  # GeoDataFrame shouldn't empty
    assert gdf.crs.to_string() == crs  # CRS is identical
    assert all(isinstance(geom, LineString) for geom in gdf.geometry)  # All geometry should LineString

    expected_number_of_geometries = 6
    assert len(gdf) == expected_number_of_geometries  # Six geometry

    # Define expected geometries
    expected_geometries = [
        LineString([(857170.91, 6856275.22, 177.24), (857192.69, 6856277.59, 177.24)]),
        LineString([(857200.29, 6856278.84, 177.24), (857228.21, 6856284.42, 177.24)]),
        LineString([(857188.43, 6856290.58, 177.93), (857219.86, 6856294.14, 177.93)]),
        LineString([(857190.83, 6856288.48, 177.93), (857220.02, 6856292.22, 177.93)]),
        LineString([(857139.26, 6856297.25, 177.25), (857174.66, 6856290.19, 177.25)]),
        LineString([(857193.85, 6856285.83, 177.25), (857225.12, 6856287.22, 177.25)]),
    ]

    # Check geometries are exactly the expected ones
    def round_linestring_coords(ls, ndigits=2):
        return LineString([(round(x, ndigits), round(y, ndigits), round(z, ndigits)) for x, y, z in ls.coords])

    actual_geoms = [round_linestring_coords(geom) for geom in gdf.geometry]

    for expected_geom in expected_geometries:
        assert any(
            expected_geom.equals(actual_geom) for actual_geom in actual_geoms
        ), f"Expected geometry not found: {expected_geom.wkt}"


def test_create_hydro_vector_mask_outside():
    # Parameters
    lines_outsides_gdf = gpd.read_file(DATA_LINES_OUTSIDE)
    bridges_gdf = gpd.read_file(DATA_BRIDGES)

    gdf = clip_lines_by_polygons(lines_outsides_gdf, bridges_gdf)

    assert gdf.empty  # GeoDataFrame should empty



import geopandas as gpd


def clip_lines_by_polygons(
    input_lines_gdf: gpd.GeoDataFrame,
    input_polygons_gdf: gpd.GeoDataFrame,
) -> gpd.GeoDataFrame:
    """Clip virtual lines by polygons (bridge deck)

    Args:
        input_lines_gdf (gpd.GeoDataFrame): virtual lines merged
        input_polygons_gdf (gpd.GeoDataFrame): Polygons (bridge deck)

    Returns:
        gpd.GeoDataFrame : all virtual lines clipped
    """
    lines_clip_gdf = gpd.overlay(input_lines_gdf, input_polygons_gdf, how="intersection")

    return lines_clip_gdf



import logging
import os
import shutil
from pathlib import Path

import geopandas as gpd
import pytest
from hydra import compose, initialize
from shapely.geometry import LineString

from las_digital_models.extract_stat_from_raster import (
    extract_z_virtual_lines_from_raster,
)

TEST_PATH = Path(__file__).resolve().parent.parent
TMP_PATH = TEST_PATH / "tmp"

DATA_DIR = TEST_PATH / "data" / "bridge"
INPUT_RASTER_DIR = DATA_DIR / "mns_hydro_postfiltre"
INPUT_GEOMETRY_DIR = DATA_DIR / "input_operators/lignes_contraintes"
INPUT_CLIP_GEOMETRY_DIR = DATA_DIR / "input_operators/tabliers"
OUTPUT_DIR = TMP_PATH / "main_extract_z_virtual_lines_from_raster"
OUTPUT_VRT_FILENAME = "MNS_HYDRO.vrt"


def setup_module():
    try:
        shutil.rmtree(TMP_PATH)

    except FileNotFoundError:
        pass
    os.mkdir(TMP_PATH)


@pytest.mark.parametrize(
    "input_raster_dir, input_geometry_dir, input_clip_geometry_dir, input_geometry_filename, input_clip_geometry_filename",
    [
        (  # Standard values
            INPUT_RASTER_DIR, INPUT_GEOMETRY_DIR, INPUT_CLIP_GEOMETRY_DIR, 
            "NUALHD_1-0_DF_lignes_contrainte.geojson", "NUALHD_1-0_DF_tabliers_pont.geojson",
        ),
        (
            # Values where we multiLineString in past version
            TEST_PATH / "data" / "data_test_linestring", TEST_PATH / "data" / "data_test_linestring/", 
            TEST_PATH / "data" / "data_test_linestring/", 
            "lignes_contraintes.geojson", "ponts.geojson"
        ),
    ],
)
def test_extract_z_virtual_lines_from_raster_default(input_raster_dir, input_geometry_dir, input_clip_geometry_dir, 
                                                     input_geometry_filename, input_clip_geometry_filename):
    output_vrt_filename = OUTPUT_VRT_FILENAME
    output_dir = OUTPUT_DIR
    output_geometry_filename = "constraint_lines.GeoJSON"
    os.makedirs(output_dir, exist_ok=True)

    with initialize(version_base="1.2", config_path="../../configs"):
        # config is relative to a module
        cfg = compose(
            config_name="config",
            overrides=[
                f"extract_stat.input_raster_dir={input_raster_dir}",
                f"extract_stat.input_geometry_dir={input_geometry_dir}",
                f"extract_stat.input_clip_geometry_dir={input_clip_geometry_dir}",
                f"extract_stat.input_geometry_filename={input_geometry_filename}",
                f"extract_stat.input_clip_geometry_filename={input_clip_geometry_filename}",
                f"extract_stat.output_vrt_filename={output_vrt_filename}",
                f"extract_stat.output_dir={output_dir}",
                f"extract_stat.output_geometry_filename={output_geometry_filename}",
            ],
        )
    extract_z_virtual_lines_from_raster.run_extract_z_virtual_lines_from_raster(cfg)
    assert (Path(output_dir) / "constraint_lines.GeoJSON").is_file()

    # Check geometry and Z value from output
    output_geometry_path = OUTPUT_DIR / output_geometry_filename
    gdf = gpd.read_file(output_geometry_path)
    assert not gdf.empty

    def is_linestring_z(geom):
        return isinstance(geom, LineString) and geom.has_z

    def all_z_coords_equal(geom):
        z_values = [coord[2] for coord in geom.coords]
        return all(z == z_values[0] for z in z_values)

    for geom in gdf.geometry:
        assert is_linestring_z(geom)  # the geometry's output is okay : LineString Z
        assert all_z_coords_equal(geom)  # this lines have the same Z value"


def test_extract_z_virtual_lines_from_raster_no_input_raster():
    input_geometry_dir = INPUT_GEOMETRY_DIR
    input_clip_geometry_dir = INPUT_CLIP_GEOMETRY_DIR
    input_geometry_filename = "NUALHD_1-0_DF_lignes_contrainte.shp"
    output_vrt_filename = OUTPUT_VRT_FILENAME
    output_dir = OUTPUT_DIR
    output_geometry_filename = "constraint_lines.GeoJSON"
    os.makedirs(output_dir, exist_ok=True)

    with initialize(version_base="1.2", config_path="../../configs"):
        # config is relative to a module
        cfg = compose(
            config_name="config",
            overrides=[
                f"extract_stat.input_geometry_dir={input_geometry_dir}",
                f"extract_stat.input_clip_geometry_dir={input_clip_geometry_dir}",
                f"extract_stat.input_geometry_filename={input_geometry_filename}",
                f"extract_stat.output_vrt_filename={output_vrt_filename}",
                f"extract_stat.output_dir={output_dir}",
                f"extract_stat.output_geometry_filename={output_geometry_filename}",
            ],
        )
    with pytest.raises(ValueError):
        extract_z_virtual_lines_from_raster.run_extract_z_virtual_lines_from_raster(cfg)


def test_extract_z_virtual_lines_from_raster_no_input_geometry():
    input_raster_dir = INPUT_RASTER_DIR
    output_vrt_filename = OUTPUT_VRT_FILENAME
    output_dir = OUTPUT_DIR
    output_geometry_filename = "constraint_lines.GeoJSON"
    os.makedirs(output_dir, exist_ok=True)

    with initialize(version_base="1.2", config_path="../../configs"):
        # config is relative to a module
        cfg = compose(
            config_name="config",
            overrides=[
                f"extract_stat.input_raster_dir={input_raster_dir}",
                f"extract_stat.output_vrt_filename={output_vrt_filename}",
                f"extract_stat.output_dir={output_dir}",
                f"extract_stat.output_geometry_filename={output_geometry_filename}",
            ],
        )
    with pytest.raises(ValueError):
        extract_z_virtual_lines_from_raster.run_extract_z_virtual_lines_from_raster(cfg)


def test_extract_z_virtual_lines_from_raster_no_output():
    input_raster_dir = INPUT_RASTER_DIR
    input_geometry_dir = INPUT_GEOMETRY_DIR
    input_clip_geometry_dir = INPUT_CLIP_GEOMETRY_DIR
    input_geometry_filename = "NUALHD_1-0_DF_lignes_contrainte.shp"
    output_vrt_filename = OUTPUT_VRT_FILENAME

    with initialize(version_base="1.2", config_path="../../configs"):
        # config is relative to a module
        cfg = compose(
            config_name="config",
            overrides=[
                f"extract_stat.input_raster_dir={input_raster_dir}",
                f"extract_stat.input_geometry_dir={input_geometry_dir}",
                f"extract_stat.input_clip_geometry_dir={input_clip_geometry_dir}",
                f"extract_stat.input_geometry_filename={input_geometry_filename}",
                f"extract_stat.output_vrt_filename={output_vrt_filename}",
                "extract_stat.output_dir=null",
                "extract_stat.output_geometry_filename=null",
            ],
        )
        with pytest.raises(ValueError, match="config.extract_stat.output_dir is empty"):
            extract_z_virtual_lines_from_raster.run_extract_z_virtual_lines_from_raster(cfg)


def test_extract_z_virtual_lines_from_raster_outside():
    input_raster_dir = INPUT_RASTER_DIR
    input_geometry_dir = INPUT_GEOMETRY_DIR
    input_clip_geometry_dir = INPUT_CLIP_GEOMETRY_DIR
    input_geometry_filename = "NUALHD_1-0_DF_lignes_contrainte_outside_MNS.geojson"
    input_clip_geometry_filename = "NUALHD_1-0_DF_tabliers_pont.geojson"
    output_vrt_filename = OUTPUT_VRT_FILENAME
    output_dir = TMP_PATH / "main_extract_z_virtual_lines_from_raster_outside"
    output_geometry_filename = "constraint_lines.GeoJSON"
    os.makedirs(output_dir, exist_ok=True)

    with initialize(version_base="1.2", config_path="../../configs"):
        # config is relative to a module
        cfg = compose(
            config_name="config",
            overrides=[
                f"extract_stat.input_raster_dir={input_raster_dir}",
                f"extract_stat.input_geometry_dir={input_geometry_dir}",
                f"extract_stat.input_clip_geometry_dir={input_clip_geometry_dir}",
                f"extract_stat.input_geometry_filename={input_geometry_filename}",
                f"extract_stat.input_clip_geometry_filename={input_clip_geometry_filename}",
                f"extract_stat.output_vrt_filename={output_vrt_filename}",
                f"extract_stat.output_dir={output_dir}",
                f"extract_stat.output_geometry_filename={output_geometry_filename}",
            ],
        )
    with pytest.raises(ValueError, match="All geometries returned None. Abort."):
        extract_z_virtual_lines_from_raster.run_extract_z_virtual_lines_from_raster(cfg)

    assert not (Path(output_dir) / "constraint_lines.GeoJSON").is_file()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    test_extract_z_virtual_lines_from_raster_default()
    test_extract_z_virtual_lines_from_raster_no_input_geometry()
    test_extract_z_virtual_lines_from_raster_no_input_raster()
    test_extract_z_virtual_lines_from_raster_no_output()



import os
import shutil
import test.utils.raster_utils as ru
from pathlib import Path

import laspy
import numpy as np
from hydra import compose, initialize

from las_digital_models.main import main_las_digital_models

COORD_X = 77055
COORD_Y = 627760
TILE_COORD_SCALE = 10
TILE_WIDTH = 50
PIXEL_SIZE = 0.5
TEST_PATH = Path(__file__).resolve().parent
TMP_PATH = TEST_PATH / "tmp" / "main"
DATA_PATH = TEST_PATH / "data"

INPUT_FILENAME = f"test_data_{COORD_X}_{COORD_Y}_LA93_IGN69.laz"
INPUT_TILENAME = os.path.splitext(INPUT_FILENAME)[0]
OUTPUT_TIF_NAME = f"{INPUT_TILENAME}_50CM.tif"

expected_xmin = COORD_X * TILE_COORD_SCALE - PIXEL_SIZE / 2
expected_ymax = COORD_Y * TILE_COORD_SCALE + PIXEL_SIZE / 2
EXPECTED_RASTER_BOUNDS = (expected_xmin, expected_ymax - TILE_WIDTH), (expected_xmin + TILE_WIDTH, expected_ymax)


def setup_module():
    try:
        shutil.rmtree(TMP_PATH)

    except FileNotFoundError:
        pass
    os.makedirs(TMP_PATH)


def get_2d_bounding_box(path):
    """Get bbox for a las file (x, y only)"""
    with laspy.open(path) as f:
        mins = f.header.mins
        maxs = f.header.maxs

    return mins[:2], maxs[:2]


def test_main_intermediate_files():
    in_mins, in_maxs = get_2d_bounding_box(DATA_PATH / INPUT_FILENAME)
    buffer_size = 10
    output_dir = TMP_PATH / "main_intermediate_files"
    output_buffer_dir = output_dir / "buffer"
    output_dtm_dir = output_dir / "DTM"
    output_dsm_dir = output_dir / "DSM"
    output_dhm_dir = output_dir / "DHM"

    with initialize(version_base="1.2", config_path="../configs"):
        # config is relative to a module
        cfg = compose(
            config_name="test",
            overrides=[
                f"io.input_filename={INPUT_FILENAME}",
                f"io.input_dir={DATA_PATH}",
                f"io.output_dir={output_dir}",
                f"tile_geometry.tile_coord_scale={TILE_COORD_SCALE}",
                f"tile_geometry.tile_width={TILE_WIDTH}",
                f"buffer.size={buffer_size}",
                "buffer.output_subdir='buffer'",
            ],
        )
        main_las_digital_models(cfg)

    # Check buffer files are correct
    output_buffer_path = output_buffer_dir / INPUT_FILENAME
    assert os.path.isfile(output_buffer_path)
    out_mins, out_maxs = get_2d_bounding_box(output_buffer_path)
    assert np.all(out_mins == in_mins - buffer_size)
    assert np.all(out_maxs[0] == in_maxs[0] + buffer_size)
    assert out_maxs[1] == in_maxs[1]  # neighbor file does not exist

    # Check output tif files are correct
    output_dtm_path = output_dtm_dir / OUTPUT_TIF_NAME
    assert os.path.isfile(output_dtm_path)
    dtm_bounds = ru.get_tif_extent(output_dtm_path)

    assert dtm_bounds == EXPECTED_RASTER_BOUNDS

    output_dsm_path = output_dsm_dir / OUTPUT_TIF_NAME
    assert os.path.isfile(output_dsm_path)
    dsm_bounds = ru.get_tif_extent(output_dsm_path)

    assert dsm_bounds == EXPECTED_RASTER_BOUNDS

    output_dhm_path = output_dhm_dir / OUTPUT_TIF_NAME
    assert os.path.isfile(output_dhm_path)
    dhm_bounds = ru.get_tif_extent(output_dhm_path)

    assert dhm_bounds == EXPECTED_RASTER_BOUNDS


def test_main_without_intermediate_files():
    """Compute only dhm, and check buffer / dsm / dtm files are only temporary files"""
    buffer_size = 10
    output_dir = TMP_PATH / "main_without_intermediate_files"
    output_buffer_dir = output_dir / "buffer"
    output_dtm_dir = output_dir / "DTM"
    output_dsm_dir = output_dir / "DSM"
    output_dhm_dir = output_dir / "DHM"
    with initialize(version_base="1.2", config_path="../configs"):
        # config is relative to a module
        cfg = compose(
            config_name="test",
            overrides=[
                f"io.input_filename={INPUT_FILENAME}",
                f"io.input_dir={DATA_PATH}",
                f"io.output_dir={output_dir}",
                f"tile_geometry.tile_coord_scale={TILE_COORD_SCALE}",
                f"tile_geometry.tile_width={TILE_WIDTH}",
                f"buffer.size={buffer_size}",
                "tasks.dtm=false",
                "tasks.dsm=false",
                "tasks.dhm=true",
            ],
        )
        main_las_digital_models(cfg)

    # Check outputs are correct
    assert not os.path.exists(output_buffer_dir)
    assert not os.path.exists(output_dtm_dir)
    assert not os.path.exists(output_dsm_dir)
    assert os.path.exists(output_dhm_dir / OUTPUT_TIF_NAME)



"""Main entry point for digital models generation"""

import logging as log
import os
import tempfile
from pathlib import Path

import hydra
from omegaconf import DictConfig
from pdaltools.las_add_buffer import create_las_with_buffer
from pdaltools.las_info import get_tile_origin_using_header_info

from las_digital_models.dhm_one_tile import run_dhm_on_tile
from las_digital_models.ip_one_tile import run_ip_on_tile


@hydra.main(config_path="../configs/", config_name="config.yaml", version_base="1.2")
def main_las_digital_models(config: DictConfig):
    """Main function to generate digital models from a las file.
    It can compute:
    - DSM: digital surface model
    - DTM: digital terrain model
    - DHM: digital height model (DSM - DTM)

    If a buffer size is set in config, the las file is buffered using its neighbors
    before generating the digital models to prevent border effects.
    Neighbors search is computed based on file names:
    las files are expected to be formatted as: {prefix1}_{prefix2}_{XXXX}_{YYYY}_{suffix}
    with XXXX and YYYY their coordinates with conditions:
    - XXXX (and YYYY) * config.tile_geometry.tile_coord_scale are the coordinates of the upper left
    corner in meters
    - XXXX (and YYYY) * config.tile_geometry.tile_coord_scale + (- for YYYY) config.tile_geometry.tile_width
    are the coordinates of the lower right corner in meters

    Args:
        config (DictConfig): hydra config for the project
    """
    log.basicConfig(level=log.INFO, format="%(message)s")

    initial_las_filename = config.io.input_filename
    in_dir = Path(config.io.input_dir)
    out_dir = Path(config.io.output_dir)

    # Check input/output files and folders
    if initial_las_filename is None or in_dir is None or out_dir is None:
        raise RuntimeError(
            """In input you have to give a las, an input directory and an output directory.
            For more info run the same command by adding --help"""
        )

    os.makedirs(out_dir, exist_ok=True)

    with (
        tempfile.TemporaryDirectory(prefix="tmp_buffer", dir=".") as tmpdir_buffer,
        tempfile.TemporaryDirectory(prefix="tmp_dtm", dir=".") as tmpdir_dtm,
        tempfile.TemporaryDirectory(prefix="tmp_dsm", dir=".") as tmpdir_dsm,
    ):
        # Get pointcloud origin from the las file metadata
        tile_origin = get_tile_origin_using_header_info(
            in_dir / initial_las_filename, tile_width=config.tile_geometry.tile_width
        )
        # Buffer
        if config.buffer.size:
            log.info(f"Create buffered las file with buffer = {config.buffer.size}")
            if config.buffer.output_subdir:
                buffer_output_dir = Path(out_dir) / config.buffer.output_subdir
            else:
                buffer_output_dir = Path(tmpdir_buffer)
            buffer_output_dir.mkdir(parents=True, exist_ok=True)

            epsg = config.io.spatial_reference
            create_las_with_buffer(
                input_dir=str(in_dir),
                tile_filename=str(in_dir / initial_las_filename),
                output_filename=str(buffer_output_dir / initial_las_filename),
                buffer_width=config.buffer.size,
                spatial_ref=f"EPSG:{epsg}" if str(epsg).isdigit() else epsg,
                tile_width=config.tile_geometry.tile_width,
                tile_coord_scale=config.tile_geometry.tile_coord_scale,
            )
        else:
            log.info("Skip las buffer creation")
            buffer_output_dir = in_dir

        # Compute DTM
        if config.tasks.dtm or config.tasks.dhm:
            log.info("Create DTM")
            dtm_output_dir = (
                os.path.join(out_dir, config.interpolation.dtm.output_subfolder) if config.tasks.dtm else tmpdir_dtm
            )
            run_ip_on_tile(
                input_dir=buffer_output_dir,
                input_filename=initial_las_filename,
                origin=tile_origin,
                output_dir=dtm_output_dir,
                pixel_size=config.tile_geometry.pixel_size,
                tile_width=config.tile_geometry.tile_width,
                spatial_reference=config.io.spatial_reference,
                no_data_value=config.tile_geometry.no_data_value,
                no_data_mask_shapefile=config.io.no_data_mask_shapefile,
                filter_dimension=config.interpolation.dtm.filter.dimension,
                filter_keep_values=config.interpolation.dtm.filter.keep_values,
            )

        # Compute DSM
        if config.tasks.dsm or config.tasks.dhm:
            log.info("Create DSM")
            dsm_output_dir = (
                os.path.join(out_dir, config.interpolation.dsm.output_subfolder) if config.tasks.dsm else tmpdir_dsm
            )
            run_ip_on_tile(
                input_dir=buffer_output_dir,
                input_filename=initial_las_filename,
                origin=tile_origin,
                output_dir=dsm_output_dir,
                pixel_size=config.tile_geometry.pixel_size,
                tile_width=config.tile_geometry.tile_width,
                spatial_reference=config.io.spatial_reference,
                no_data_value=config.tile_geometry.no_data_value,
                no_data_mask_shapefile=config.io.no_data_mask_shapefile,
                filter_dimension=config.interpolation.dsm.filter.dimension,
                filter_keep_values=config.interpolation.dsm.filter.keep_values,
            )

        # Compute DHM
        if config.tasks.dhm:
            log.info("Create DHM")
            dhm_output_dir = os.path.join(out_dir, config.dhm.output_subfolder)
            run_dhm_on_tile(
                input_las_filename=initial_las_filename,
                input_dtm_dir=dtm_output_dir,
                input_dsm_dir=dsm_output_dir,
                output_dir=dhm_output_dir,
                pixel_size=config.tile_geometry.pixel_size,
                no_data_value=config.tile_geometry.no_data_value,
            )


if __name__ == "__main__":
    main_las_digital_models()



"""
Main script to run the extraction of minimum Z values along lines (defined in a geometry file) \
from raster containing Z value
"""

import logging
import os

import geopandas as gpd
import hydra
from omegaconf import DictConfig
from osgeo import gdal

from las_digital_models.commons import commons
from las_digital_models.extract_stat_from_raster.rasters.extract_z_min_from_raster_by_polylines import (
    clip_lines_by_raster,
    extract_polylines_min_z_from_dsm,
)
from las_digital_models.extract_stat_from_raster.vectors.clip_geometry import (
    clip_lines_by_polygons,
)

gdal.UseExceptions()

log = commons.get_logger(__name__)


def create_vrt(dir_list_raster: list, output_vrt: str):
    """Create vrt from raster files in a directory.

    Args:
        dir_list_raster (List): ist of input raster files.
        output_vrt (str): Path to the output VRT file.

    Raises:
        ValueError: If the VRT doesn't create
    """
    # Build and save VRT file
    vrt_options = gdal.BuildVRTOptions(resampleAlg="cubic", addAlpha=True)
    my_vrt = gdal.BuildVRT(output_vrt, dir_list_raster, options=vrt_options)

    if my_vrt is None:
        raise ValueError(f"gdal.BuildVRT returned None for {output_vrt}")

    my_vrt = None  # necessary to close the VRT file properly


@hydra.main(config_path="../../configs/", config_name="config.yaml", version_base="1.2")
def run_extract_z_virtual_lines_from_raster(config: DictConfig):
    """Extract the minimum Z value along one or more 2d lines (contained in a geometry file) using hydra config
    config parameters are explained in the default.yaml files

    Raises:
        RuntimeError: If the input RASTER file has no valid EPSG code.
        ValueError: if the geometry file does not only contain (Multi)LineStrings.
    """
    # Check input files
    raster_dir = config.extract_stat.input_raster_dir
    if not os.path.isdir(raster_dir):
        raise ValueError(
            f"""config.extract_stat.input_raster_dir ({config.extract_stat.input_raster_dir}) not found"""
        )

    input_geometry = os.path.join(config.extract_stat.input_geometry_dir, config.extract_stat.input_geometry_filename)

    if not os.path.isfile(input_geometry):
        raise ValueError(f"Input geometry file not found: {input_geometry}")

    dir_list_raster = [os.path.join(raster_dir, f) for f in os.listdir(raster_dir) if f.lower().endswith(".tif")]
    if not dir_list_raster:
        raise ValueError(f"No raster (.tif) files found in {raster_dir}")

    input_clip_geometry = os.path.join(
        config.extract_stat.input_clip_geometry_dir, config.extract_stat.input_clip_geometry_filename
    )
    # path to the geometry file
    if not os.path.isfile(input_geometry):
        raise ValueError(f"Input geometry file not found: {input_clip_geometry}")

    # Check output folder
    output_dir = config.extract_stat.output_dir
    if output_dir is None:
        raise ValueError(
            """config.extract_stat.output_dir is empty, please provide an output directory in the configuration"""
        )
    os.makedirs(config.extract_stat.output_dir, exist_ok=True)

    # Parameters
    spatial_ref = config.extract_stat.spatial_reference
    output_geometry = os.path.join(config.extract_stat.output_dir, config.extract_stat.output_geometry_filename)
    output_vrt = os.path.join(config.extract_stat.output_dir, config.extract_stat.output_vrt_filename)

    # Create  vrt
    create_vrt(dir_list_raster, output_vrt)

    # Read the input GeoJSON
    geom_gdf = gpd.read_file(input_geometry)
    polygons_gdf = gpd.read_file(input_clip_geometry)

    # Convert geometries to LineString (no more MultiLineString)
    mask = geom_gdf.geometry.geom_type.isin(['LineString', 'MultiLineString'])
    lines_gdf = geom_gdf.loc[mask].explode(index_parts=False).reset_index(drop=True)

    if geom_gdf.crs is None:
        geom_gdf.set_crs(epsg=spatial_ref, inplace=True)

    if polygons_gdf.crs is None:
        polygons_gdf.set_crs(epsg=spatial_ref, inplace=True)

    # Keep lines inside raster (VRT created)
    lines_gdf_clip = clip_lines_by_raster(lines_gdf, output_vrt, spatial_ref)

    # Extract Z value from lines and clean the result
    lines_gdf_min_z = (
        extract_polylines_min_z_from_dsm(lines_gdf_clip, output_vrt, no_data_value=config.tile_geometry.no_data_value)
        .drop(columns=[c for c in ["index", "FID"] if c in lines_gdf.columns], errors="ignore")
        .reset_index(drop=True)
    )

    # Check lines are not empty
    if lines_gdf_min_z.empty:
        raise ValueError("All geometries returned None. Abort.")

    # Clip lines by bridges
    geoms_gdf_min_z_clip = clip_lines_by_polygons(lines_gdf_min_z, polygons_gdf)
    
    # Convert geometries to LineString (no more MultiLineString)
    mask = geoms_gdf_min_z_clip.geometry.geom_type.isin(['LineString', 'MultiLineString'])
    lines_gdf_min_z_clip = geoms_gdf_min_z_clip.loc[mask].explode(index_parts=False).reset_index(drop=True)

    lines_gdf_min_z_clip.to_file(output_geometry, driver="GeoJSON")


def main():
    logging.basicConfig(level=logging.INFO)
    run_extract_z_virtual_lines_from_raster()


if __name__ == "__main__":
    main()



import logging
import os
import shutil
import test.utils.raster_utils as ru
from pathlib import Path

from hydra import compose, initialize

from las_digital_models import ip_one_tile

COORD_X = 77055
COORD_Y = 627760
TILE_COORD_SCALE = 10
TILE_WIDTH = 50
PIXEL_SIZE = 0.5

TEST_PATH = Path(__file__).resolve().parent
TMP_PATH = TEST_PATH / "tmp" / "ip_one_tile"
GROUND_TRUTH_FOLDER = TEST_PATH / "data" / "interpolation"

EXPECTED_XMIN = COORD_X * TILE_COORD_SCALE - PIXEL_SIZE / 2
EXPECTED_YMAX = COORD_Y * TILE_COORD_SCALE + PIXEL_SIZE / 2
EXPECTED_RASTER_BOUNDS = (EXPECTED_XMIN, EXPECTED_YMAX - TILE_WIDTH), (EXPECTED_XMIN + TILE_WIDTH, EXPECTED_YMAX)

SHAPEFILE = TEST_PATH / "data" / "mask_shapefile" / "test_multipolygon_shapefile.shp"
EXPECTED_OUTPUT_USING_SHAPEFILE = GROUND_TRUTH_FOLDER / "test_data_77055_627760_LA93_IGN69_50CM_no_data.tif"


def setup_module():
    try:
        shutil.rmtree(TMP_PATH)

    except FileNotFoundError:
        pass
    os.mkdir(TMP_PATH)


def get_expected_output_file(base_dir=None):
    if base_dir is None:
        base_dir = TMP_PATH / "hydra_ip"
    expected_output_file = os.path.join(base_dir, f"test_data_{COORD_X}_{COORD_Y}_LA93_IGN69_50CM.tif")

    return expected_output_file


def test_main():
    output_dir = os.path.join(TMP_PATH, "test_main")
    os.makedirs(output_dir, exist_ok=True)
    with initialize(version_base="1.2", config_path="../configs"):
        # config is relative to a module
        cfg = compose(
            config_name="config",
            overrides=[
                "io=test",
                f"io.output_dir={output_dir}",
                "tile_geometry=test",
                "interpolation.custom.filter.dimension=''",
                "interpolation.custom.filter.keep_values=[]",
            ],
        )
        output_file = get_expected_output_file(base_dir=output_dir)
        logging.debug(output_file)
        logging.debug(f"Pixel size: {cfg.tile_geometry.pixel_size}")

        ip_one_tile.main(cfg)
        assert os.path.isfile(output_file)

        raster_bounds = ru.get_tif_extent(output_file)
        assert ru.allclose_mm(raster_bounds, EXPECTED_RASTER_BOUNDS)

        assert ru.tif_values_all_close(output_file, os.path.join(GROUND_TRUTH_FOLDER, os.path.basename(output_file)))


def test_main_with_no_data_mask():
    output_dir = os.path.join(TMP_PATH, "test_main_with_no_data_mask")
    os.makedirs(output_dir, exist_ok=True)
    with initialize(version_base="1.2", config_path="../configs"):
        # config is relative to a module
        cfg = compose(
            config_name="config",
            overrides=[
                "io=test",
                "tile_geometry=test",
                f"io.no_data_mask_shapefile={SHAPEFILE}",
                f"io.output_dir={output_dir}",
                "interpolation.custom.filter.keep_values=[]",
            ],
        )

        output_file = get_expected_output_file(base_dir=output_dir)
        logging.debug(f"Write to {output_file}")

        ip_one_tile.main(cfg)
        assert os.path.isfile(output_file)

        raster_bounds = ru.get_tif_extent(output_file)
        assert ru.allclose_mm(raster_bounds, EXPECTED_RASTER_BOUNDS)

        assert ru.tif_values_all_close(output_file, EXPECTED_OUTPUT_USING_SHAPEFILE)



"""Run create DHM on a single tile (current definition is DSM - DTM)"""

import logging
import os

import hydra
from omegaconf import DictConfig

from las_digital_models.commons import commons
from las_digital_models.tasks.dhm_generation import calculate_dhm

log = commons.get_logger(__name__)


def run_dhm_on_tile(
    input_las_filename: str,
    input_dtm_dir: str,
    input_dsm_dir: str,
    output_dir: str,
    pixel_size: float,
    no_data_value: float,
):
    os.makedirs(output_dir, exist_ok=True)
    tilename, _ = os.path.splitext(input_las_filename)

    # for export
    _size = commons.give_name_resolution_raster(pixel_size)
    geotiff_filename = f"{tilename}{_size}.tif"
    geotiff_dsm = os.path.join(input_dsm_dir, geotiff_filename)
    geotiff_dtm = os.path.join(input_dtm_dir, geotiff_filename)
    geotiff_output = os.path.join(output_dir, geotiff_filename)
    # process
    calculate_dhm(geotiff_dsm, geotiff_dtm, geotiff_output, no_data_value=no_data_value)

    return


@hydra.main(config_path="../configs/", config_name="config.yaml", version_base="1.2")
def main(config: DictConfig):
    logging.basicConfig(level=logging.INFO)

    run_dhm_on_tile(
        input_las_filename=config.io.input_filename,
        input_dtm_dir=config.dhm.input_dtm_dir,
        input_dsm_dir=config.dhm.input_dsm_dir,
        output_dir=config.io.output_dir,
        pixel_size=config.tile_geometry.pixel_size,
        no_data_value=config.tile_geometry.no_data_value,
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# maintener : MDupays
# version : v.0.1.0 02/02/2023
# Calculate DHM
from osgeo import gdal
from osgeo_utils import gdal_calc

gdal.UseExceptions()


def calculate_dhm(input_image_dsm: str, input_image_dtm: str, output_image: str, no_data_value: int = -9999):
    """Calculate DHM from DSM and DTM (DHM = DSM - DTM)

    Args:
        input_file_dsm (str): path to DSM file
        input_file_dtm (str): path to DTM file
        output_image (str) : path to output DHM file
        no_data_value (int): no data value (default to -9999)

    """
    # Calculate A - B where A and B have valid values, else return a no_data value
    gdal_calc.Calc(
        f"(A - B) * (A != {no_data_value}) * (B != {no_data_value})"
        + f" + ({no_data_value}) * (1 - (A != {no_data_value}) * (B != {no_data_value}))",
        outfile=output_image,
        A=input_image_dsm,
        A_band=1,
        B=input_image_dtm,
        B_band=1,
        type="Float32",
        quiet=True,
        NoDataValue=no_data_value,
        overwrite=True,
    )



import logging
import os
import shutil
import subprocess as sp

import pytest

from las_digital_models.commons import commons

test_path = os.path.dirname(os.path.abspath(__file__))
tmp_path = os.path.join(test_path, "tmp", "run_script")
input_dir = os.path.join(test_path, "data")
output_dir = tmp_path
file_ext = "laz"
pixel_size = 0.5

expected_output_dirs = {
    "dtm": os.path.join(output_dir, "DTM"),
    "dsm": os.path.join(output_dir, "DSM"),
    "dhm": os.path.join(output_dir, "DHM"),
}


def setup_module(module):
    try:
        shutil.rmtree(tmp_path)

    except FileNotFoundError:
        pass
    os.makedirs(tmp_path)


@pytest.mark.functional_test
def test_run_script():
    cmd = ["./run.sh", "-i", input_dir, "-o", output_dir, "-p", str(pixel_size), "-c", "test"]
    print(cmd)
    r = sp.run(cmd, capture_output=True)
    logging.debug(f"Stdout is: {r.stdout.decode()}")
    logging.debug(f"Stderr is: {r.stderr.decode()}")
    if r.returncode == 1:
        msg = r.stderr.decode()

        pytest.fail(f"Test for run.sh failed with message: {msg}", True)

    # Check that all files are created (for all methods)

    for input_file in os.listdir(input_dir):
        if input_file.endswith(("las", "laz")):
            tilename = os.path.splitext(input_file)[0]
            for od in expected_output_dirs.keys():
                _size = commons.give_name_resolution_raster(pixel_size)
                out_filename = f"{tilename}{_size}.tif"
                out_path = os.path.join(expected_output_dirs[od], out_filename)
                assert os.path.isfile(out_path), f"Output for {od} was not generated"


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    test_run_script()



"""Main script for interpolation on a single tile
Output files will be written to the target folder, tagged with the name of the interpolation
method that was used.
"""

import logging
import os
import tempfile
from typing import List, Tuple

import hydra
from omegaconf import DictConfig
from pdaltools.las_info import parse_filename

from las_digital_models.commons import commons
from las_digital_models.tasks.las_interpolation import interpolate
from las_digital_models.tasks.postprocessing import mask_with_no_data_shapefile

log = commons.get_logger(__name__)


def run_ip_on_tile(
    input_dir: str,
    input_filename: str,
    output_dir: str,
    origin: Tuple[int, int],
    pixel_size: float,
    tile_width: int,
    spatial_reference: str,
    no_data_value: float,
    no_data_mask_shapefile: str,
    filter_dimension: str,
    filter_keep_values: List[int],
):
    """Run interpolation on single tile with geometry masking in case a mask is provided"""

    # Generate output filenames
    tilename, _ = os.path.splitext(input_filename)
    _size = commons.give_name_resolution_raster(pixel_size)
    geotiff_stem = f"{tilename}{_size}"
    geotiff_filename = f"{geotiff_stem}.tif"
    geotiff_path = os.path.join(output_dir, geotiff_filename)

    input_path = os.path.join(input_dir, input_filename)
    os.makedirs(output_dir, exist_ok=True)

    if no_data_mask_shapefile:
        with tempfile.NamedTemporaryFile(suffix=".tif", prefix=f"{geotiff_stem}_raw") as tmp_geotiff:
            # process interpolation
            interpolate(
                input_path,
                tmp_geotiff.name,
                origin,
                pixel_size,
                tile_width,
                spatial_reference,
                no_data_value,
                filter_dimension,
                filter_keep_values,
            )
            mask_with_no_data_shapefile(no_data_mask_shapefile, tmp_geotiff.name, geotiff_path, no_data_value)

    else:
        interpolate(
            input_path,
            geotiff_path,
            origin,
            pixel_size,
            tile_width,
            spatial_reference,
            no_data_value,
            filter_dimension,
            filter_keep_values,
        )


@hydra.main(config_path="../configs/", config_name="config.yaml", version_base="1.2")
def main(config: DictConfig):
    logging.basicConfig(level=logging.INFO)

    # Use filename to get the tile coordinates.
    # Coordinates are needed to define neighboring tiles, in order to create the buffered tile.
    input_path = os.path.join(config.io.input_dir, config.io.input_filename)
    _, coordX, coordY, _ = parse_filename(input_path)
    origin = [
        float(coordX) * config.tile_geometry.tile_coord_scale,
        float(coordY) * config.tile_geometry.tile_coord_scale,
    ]

    run_ip_on_tile(
        input_dir=config.io.input_dir,
        input_filename=config.io.input_filename,
        origin=origin,
        output_dir=config.io.output_dir,
        pixel_size=config.tile_geometry.pixel_size,
        tile_width=config.tile_geometry.tile_width,
        spatial_reference=config.io.spatial_reference,
        no_data_value=config.tile_geometry.no_data_value,
        no_data_mask_shapefile=config.io.no_data_mask_shapefile,
        filter_dimension=config.interpolation.custom.filter.dimension,
        filter_keep_values=config.interpolation.custom.filter.keep_values,
    )


if __name__ == "__main__":
    main()



import logging

import geopandas as gpd
import rasterio
from rasterstats import zonal_stats
from shapely.geometry import LineString, box


def clip_lines_by_raster(input_lines: gpd.GeoDataFrame, input_raster: str, crs: str = None) -> gpd.GeoDataFrame:
    """
    Select lines from a GeoDataFrame that intersect the raster extent.

    Args:
        input_lines (gpd.GeoDataFrame): GeoDataFrame with lines.
        input_raster (str): Path to the raster file (e.g., .tif).
        crs (str, optional): Target CRS if reprojection is needed.

    Returns:
        gpd.GeoDataFrame: Lines that intersect with the raster extent.
    """
    # Open the raster and get its bounding box
    with rasterio.open(input_raster) as src:
        bounds = src.bounds
        raster_crs = src.crs

    # Convert lines to raster CRS if needed
    input_lines = input_lines.to_crs(crs if crs else raster_crs)
    bbox_polygon = box(*bounds)

    # Select lines that intersect the raster bbox
    clipped_lines = input_lines[input_lines.intersects(bbox_polygon)]

    return clipped_lines


def extract_polylines_min_z_from_dsm(
    lines_gdf: gpd.GeoDataFrame, dsm_rasterpath: str, no_data_value: int = 9999
) -> gpd.GeoDataFrame:
    """
    Extracts the minimum Z value from a DSM raster for each polyline (LineString or MultiLineString)
    in the input shapefile, keeping the original geometry.

    Args:
        lines_gdf (str): GeoDataFrame with 2D lines.
        dsm_rasterpath (str): Path to the DSM raster (.vrt).
        no_data_value (int): no data value (default to -9999)

    Returns:
        GeoDataFrame: A GeoDataFrame with generated 3D Lines.
    """

    def get_z_min_on_linestring(geom, dsm_rasterpath):
        if isinstance(geom, LineString):
            stats = zonal_stats(
                vectors=[geom], raster=dsm_rasterpath, stats=["min"], all_touched=True, nodata=no_data_value
            )
            min_z = stats[0]["min"]

            if min_z is None or int(min_z) == no_data_value:
                logging.warning(f"No valid Zmin found for geometry {geom} (ignored).")
                return None

            coords_3d = [(x, y, round(min_z, 2)) for x, y in geom.coords]
            return LineString(coords_3d)

        else:
            logging.warning(f"Geometry {geom} is not a LineString (ignored).")
            return None

    # Apply this function "get_z_min in linestring"
    lines_gdf["geometry"] = lines_gdf["geometry"].apply(lambda geom: get_z_min_on_linestring(geom, dsm_rasterpath))

    def is_invalid(geom):
        if geom is None:
            return True
        if isinstance(geom, LineString) and geom.has_z:
            z_vals = [pt[2] for pt in geom.coords]
            return all(z == no_data_value for z in z_vals)
        return False

    lines_gdf = lines_gdf[~lines_gdf["geometry"].apply(is_invalid)]

    return lines_gdf



import fiona
import rasterio
import rasterio.mask


def mask_with_no_data_shapefile(shapefile: str, input_raster: str, output_raster: str, no_data: int):
    """Burn no-data value inside polygons from shapefile (overwrites input raster)"""
    # # Create raster
    # out = gdal.Rasterize(raster_file, shapefile, allTouched=True, burnValues=no_data)
    # print(out)
    with fiona.open(shapefile, "r") as fhandle:
        shapes = [feature["geometry"] for feature in fhandle]

    with rasterio.open(input_raster) as src:
        out_image, _ = rasterio.mask.mask(src, shapes, crop=False, all_touched=True, invert=True)
        out_meta = src.meta

    with rasterio.open(output_raster, "w", **out_meta) as dest:
        dest.write(out_image)



"""Script to generate small test tiles from a laz file
Take a grid of 3x3 tiles of size 'tiles_size', with a given offset from the bottom-left corner of
the input tile.
Save only part of the tiles, the indices of the output tiles are located like:
6, 7, 8
3, 4, 5
0, 1, 2
(the up-left corner of tile 4 is centered to make it easier to compute)
"""

import argparse
import os
import test.utils.point_cloud_utils as pcu

import numpy as np
from las_stitching.las_clip import las_crop


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", "-i", help="Input laz file")
    parser.add_argument("--output", "-o", help="Folder for las outpout", default="test/data/")
    parser.add_argument("--tiles_size", "-s", type=int, help="Size (in meter) for the output tiles", default=10)
    parser.add_argument("--keep", "-k", type=int, help="Number of tiles to keep (out of 9). Default: all", default=9)
    parser.add_argument(
        "--offset",
        "-d",
        nargs=2,
        type=int,
        help="Offset in xy of the input tile from the corner of the input file (in meters)",
        default=[0, 0],
    )

    return parser.parse_args()


def main(args):
    bounds = pcu.get_2d_bounding_box(args.input)
    assert np.all(
        bounds[1] - bounds[0] > (np.array(args.offset) + 3 * args.tiles_size)
    ), "Output tiles fall outside of input tile"
    [xmin_in, ymin_in], _ = bounds
    gridy, gridx = np.mgrid[0:3, 0:3]
    ii = 0
    for xx, yy in zip(np.nditer(gridx), np.nditer(gridy)):
        output_path = os.path.join(args.output, f"test_data_{xx:04d}_{yy:04d}_LA93_IGN69.laz")
        xmin = xmin_in + args.offset[0] + xx * args.tiles_size
        xmax = xmin_in + args.offset[0] + (xx + 1) * args.tiles_size
        ymin = ymin_in + args.offset[1] + yy * args.tiles_size
        ymax = ymin_in + args.offset[1] + (yy + 1) * args.tiles_size
        las_crop(args.input, output_path, ([xmin, xmax], [ymin, ymax]))
        ii += 1
        if ii >= args.keep:
            break


if __name__ == "__main__":
    args = parse_args()
    main(args)



import os
import shutil
import test.utils.raster_utils as ru
from pathlib import Path

import pytest

from las_digital_models.tasks.las_interpolation import interpolate

TILE_COORD_SCALE = 10
TILE_WIDTH = 50
PIXEL_SIZE = 0.5

TEST_PATH = Path(__file__).resolve().parent.parent
TMP_PATH = TEST_PATH / "tmp" / "tasks" / "las_interpolation"
INPUT_FILE = TEST_PATH / "data" / "test_data_77055_627760_LA93_IGN69.laz"
GROUND_TRUTH_FOLDER = TEST_PATH / "data" / "interpolation"

COORD_X = 77055
COORD_Y = 627760
ORIGIN = [COORD_X * TILE_COORD_SCALE, COORD_Y * TILE_COORD_SCALE]
EXPECTED_XMIN = ORIGIN[0] - PIXEL_SIZE / 2
EXPECTED_YMAX = ORIGIN[1] + PIXEL_SIZE / 2
EXPECTED_RASTER_BOUNDS = (EXPECTED_XMIN, EXPECTED_YMAX - TILE_WIDTH), (EXPECTED_XMIN + TILE_WIDTH, EXPECTED_YMAX)


def setup_module():
    try:
        shutil.rmtree(TMP_PATH)

    except FileNotFoundError:
        pass
    os.makedirs(TMP_PATH)


@pytest.mark.parametrize(
    "filter_dimension, filter_values, output_file, ground_truth_file",
    [
        # default with no pre-filtering
        (
            "",
            [],
            TMP_PATH / "interpolate_default.tif",
            os.path.join(GROUND_TRUTH_FOLDER, "test_data_77055_627760_LA93_IGN69_50CM.tif"),
        ),
        # Filter by class (ground only)
        (
            "Classification",
            [2, 9, 66],
            TMP_PATH / "interpolate_classif.tif",
            os.path.join(GROUND_TRUTH_FOLDER, "test_data_77055_627760_LA93_IGN69_50CM_dtm_classes.tif"),
        ),
        # Filter by return number (check that other dimensions are ok)
        (
            "ReturnNumber",
            [2, 3, 4, 5],
            TMP_PATH / "interpolate_returnnumber.tif",
            os.path.join(GROUND_TRUTH_FOLDER, "test_data_77055_627760_LA93_IGN69_50CM_filter_returnnumber.tif"),
        ),
    ],
)
def test_interpolate(filter_dimension, filter_values, output_file, ground_truth_file):
    interpolate(
        INPUT_FILE,
        output_file,
        tile_origin=ORIGIN,
        pixel_size=PIXEL_SIZE,
        tile_width=TILE_WIDTH,
        spatial_ref="EPSG:2154",
        no_data_value=-9999,
        filter_dimension=filter_dimension,
        filter_values=filter_values,
    )

    assert os.path.isfile(output_file)

    raster_bounds = ru.get_tif_extent(output_file)
    assert ru.allclose_mm(raster_bounds, EXPECTED_RASTER_BOUNDS)

    assert ru.tif_values_all_close(output_file, ground_truth_file)



import os
from pathlib import Path

import geopandas as gpd
import pytest
import rasterio
from shapely.geometry import LineString

from las_digital_models.extract_stat_from_raster.rasters.extract_z_min_from_raster_by_polylines import (
    clip_lines_by_raster,
    extract_polylines_min_z_from_dsm,
)

TEST_PATH = Path(__file__).resolve().parent.parent
DATA_RASTER_PATH = os.path.join(TEST_PATH, "data/bridge/mns_hydro_postfiltre")
INPUT_RASTER = os.path.join(DATA_RASTER_PATH, "test_mns_hydro_2023_0299_6802_LA93_IGN69_5m.tif")


def test_extract_polylines_min_z_from_dsm():
    lines = [
        LineString([(299934.06, 6801817.96), (300007.48, 6801822.18)]),
        LineString([(299922.47, 6801805.18), (299993.08, 6801808.32)]),  # Line outside from raster
    ]
    lines_gdf = gpd.GeoDataFrame(geometry=lines, crs="EPSG:2154")

    result = extract_polylines_min_z_from_dsm(lines_gdf, INPUT_RASTER)

    # # Check that all resulting geometries are LineStrings
    assert all(isinstance(geom, LineString) for geom in result.geometry)
    assert all(geom.has_z for geom in result.geometry)

    # Expected minimum Z values for each line, manually determined
    expected_min_z = [
        140.91,  # first line
        140.45,  # second line
    ]

    # Check that the computed min_z matches the expected ones
    for geom, expected_z in zip(result.geometry, expected_min_z):
        zs = [round(z, 2) for x, y, z in geom.coords]
        assert all(z == expected_z for z in zs), f"Expected all Z = {expected_z}, got {zs}"


def test_line_outside_raster():
    line = LineString([(0, 0), (1, 1)])  # no intersection with raster footprint
    lines_gdf = gpd.GeoDataFrame(geometry=[line], crs="EPSG:2154")

    result = extract_polylines_min_z_from_dsm(lines_gdf, INPUT_RASTER)

    assert len(result) == 0


@pytest.fixture(scope="module")
def raster_bounds():
    with rasterio.open(INPUT_RASTER) as src:
        return src.bounds  # left, bottom, right, top


@pytest.mark.parametrize(
    "line_type, line_geom, should_be_retained",
    [
        ("inside_line", lambda b: LineString([(b.left + 10, b.bottom + 10), (b.left + 50, b.bottom + 50)]), True),
        (
            "outside_line",
            lambda b: LineString([(b.left - 1000, b.bottom - 1000), (b.left - 900, b.bottom - 900)]),
            False,
        ),
        ("edge_line", lambda b: LineString([(b.left, b.bottom + 10), (b.left, b.top - 10)]), True),
        ("touching_line", lambda b: LineString([(b.right - 1, b.top - 1), (b.right + 500, b.top + 500)]), True),
        (
            "crossing_line",
            lambda b: LineString([(b.left - 50, (b.bottom + b.top) / 2), (b.right + 50, (b.bottom + b.top) / 2)]),
            True,
        ),
    ],
)
def test_clip_lines_parametrized(raster_bounds, line_type, line_geom, should_be_retained):
    line = line_geom(raster_bounds)
    lines_gdf = gpd.GeoDataFrame(geometry=[line], crs="EPSG:2154")

    clipped = clip_lines_by_raster(lines_gdf, INPUT_RASTER)

    if should_be_retained:
        assert not clipped.empty, f"{line_type} should have been retained"
        assert clipped.geometry.intersects(line).any(), f"{line_type} not properly clipped"
    else:
        assert clipped.empty, f"{line_type} should NOT have been retained"



import os
import shutil
import test.utils.raster_utils as ru

from hydra import compose, initialize

from las_digital_models import dhm_one_tile

coordX = 77055
coordY = 627760
tile_coord_scale = 10
tile_width = 50
pixel_size = 0.5

test_path = os.path.dirname(os.path.abspath(__file__))
tmp_path = os.path.join(test_path, "tmp")

output_dir = os.path.join(tmp_path, "DHM")
expected_output_file = os.path.join(output_dir, f"test_data_{coordX}_{coordY}_LA93_IGN69_50CM.tif")

expected_xmin = coordX * tile_coord_scale - pixel_size / 2
expected_ymax = coordY * tile_coord_scale + pixel_size / 2
expected_raster_bounds = (expected_xmin, expected_ymax - tile_width), (expected_xmin + tile_width, expected_ymax)


def setup_module(module):
    try:
        shutil.rmtree(tmp_path)

    except FileNotFoundError:
        pass
    os.mkdir(tmp_path)


def test_dhm_one_tile():
    with initialize(version_base="1.2", config_path="../configs"):
        # config is relative to a module
        cfg = compose(
            config_name="config",
            overrides=[
                "io=test",
                "tile_geometry=test",
                "dhm=test",
                f"io.output_dir={output_dir}",
            ],
        )

    dhm_one_tile.main(cfg)
    assert os.path.isfile(expected_output_file)

    raster_bounds = ru.get_tif_extent(expected_output_file)
    assert ru.allclose_mm(raster_bounds, expected_raster_bounds)



from typing import List, Tuple

import pdal
from osgeo import gdal

from las_digital_models.commons import commons

gdal.UseExceptions()


@commons.eval_time_with_pid
def interpolate(
    input_file: str,
    output_file: str,
    tile_origin: Tuple[int, int],
    pixel_size: float,
    tile_width: int,
    spatial_ref: str,
    no_data_value: int,
    filter_dimension: str,
    filter_values: List[int],
):
    """Generate a Z (height) raster file from a LAS point cloud file by interpolating the Z value at the center of
    each pixel.

    It uses TIN interpolation.

    Steps are:
    - filter the points to use in the interplation (using one dimension name and a list of values)
    (eg. Classification=2(ground) for a digital terrain model)
    - triangulate the point cloud using Delaunay
    - interpolate the height values at the center of the pixels using Faceraster
    - write the result in a raster file.

    Args:
        input_file (str): path to the las/laz file to interpolate
        output_file (str): path to the output raster
        tile_origin (Tuple[int, int]): coordinates of the upper left corner of the tile (in meters)
        pixel_size (float): pixel size of the output raster in meters (pixels are supposed to be squares)
        tile_width (int): width of the tile in meters (used to infer the lower-left corner)
        spatial_ref (str): spatial reference to use when reading las file
        no_data_value (int): no data value for the output raster
        filter_dimension (str): Name of the dimension along which to filter input points
        (keep empty to disable input filter)
        filter_values (List[int]): Values to keep for input points along filter_dimension
    """

    nb_pixels = [int(tile_width / pixel_size), int(tile_width / pixel_size)]

    # Read with pdal
    pipeline = pdal.Reader.las(filename=input_file, override_srs=spatial_ref, nosrs=True)
    if filter_dimension and filter_values:
        pipeline |= pdal.Filter.range(limits=",".join(f"{filter_dimension}[{v}:{v}]" for v in filter_values))

    # Add interpolation method to the pdal pipeline
    pipeline |= pdal.Filter.delaunay()

    pipeline |= pdal.Filter.faceraster(
        resolution=str(pixel_size),
        origin_x=str(tile_origin[0] - pixel_size / 2),  # lower left corner
        origin_y=str(tile_origin[1] + pixel_size / 2 - tile_width),  # lower left corner
        width=str(nb_pixels[0]),
        height=str(nb_pixels[1]),
    )
    pipeline |= pdal.Writer.raster(gdaldriver="GTiff", nodata=no_data_value, data_type="float32", filename=output_file)

    pipeline.execute()



import numpy as np
import rasterio
from osgeo import gdal

# Make gdal forward exceptions to python pinding
gdal.UseExceptions()


# https://gis.stackexchange.com/questions/57834/how-to-get-raster-corner-coordinates-using-python-gdal-bindings
def get_tif_extent(filename):
    """Return list of corner coordinates from a tif image"""
    ds = gdal.Open(filename)

    xmin, xpixel, _, ymax, _, ypixel = ds.GetGeoTransform()
    width, height = ds.RasterXSize, ds.RasterYSize
    xmax = xmin + width * xpixel
    ymin = ymax + height * ypixel

    # close gdal dataset (cf. https://gis.stackexchange.com/questions/80366/why-close-a-dataset-in-gdal-python)
    ds = None

    return (xmin, ymin), (xmax, ymax)


def allclose_mm(a, b):
    """Check that values are similar with milimeter precision
    Use this nstead of np.allclose to use only an absolute tolerance"""
    if isinstance(a, tuple) or isinstance(b, tuple):
        a = np.array(a)
        b = np.array(b)
    return np.all(np.less(np.abs(b - a), 1e-3))


def tif_values_all_close(filename1, filename2):
    with rasterio.Env():
        src1 = rasterio.open(filename1)
        data1 = src1.read(1)

        src2 = rasterio.open(filename2)
        data2 = src2.read(1)

    return allclose_mm(data1, data2)



# -*- coding: utf-8 -*-
# maintener : MDupays
# version : v.1 06/12/2022
# COMMONS
import logging
import os
import sys
import time
from typing import Callable


def get_logger(name):
    log = logging.getLogger(__name__)
    log.setLevel(logging.INFO)
    streamHandler = logging.StreamHandler(sys.stdout)
    streamHandler.setLevel(logging.INFO)
    log.addHandler(streamHandler)

    return log


def eval_time(function: Callable):
    """decorator to log the duration of the decorated method"""

    def timed(*args, **kwargs):
        time_start = time.time()
        result = function(*args, **kwargs)
        time_elapsed = round(time.time() - time_start, 2)

        logging.info(f"Processing time of {function.__name__}: {time_elapsed}s")
        return result

    return timed


def eval_time_with_pid(function: Callable):
    """decorator to log the duration of the decorated method"""

    def timed(*args, **kwargs):
        logging.info(f"Starting {function.__name__} with PID {os.getpid()}.")
        time_start = time.time()
        result = function(*args, **kwargs)
        time_elapsed = round(time.time() - time_start, 2)
        logging.info(f"{function.__name__} with PID {os.getpid()} finished.")
        logging.info(f"Processing time of {function.__name__}: {time_elapsed}s")
        return result

    return timed


def give_name_resolution_raster(size):
    """
    Give a resolution from raster

    Args:
        size (int): raster cell size

    Return:
        _size(str): resolution from raster for output's name
    """
    size_cm = size * 100
    if int(size) == float(size):
        _size = f"_{int(size)}M"
    elif int(size_cm) == float(size_cm):
        _size = f"_{int(size_cm)}CM"
    else:
        raise ValueError(
            f"Cell size is subcentimetric ({size}m) i.e raster resolution is "
            + "too high : output name not implemented for this case"
        )

    return _size
